*******************************************************************************
*									      *
*	       Motorola 680x0 fixedpoint math library v1.4		      *
*									      *
* Assembles with hisoft devpac						      *
* by ray//.tscc	     2002-2005						      *
*									      *
*******************************************************************************
* www: http://ray.tscc.de						      *
* eml: ray.tscc.de							      *
*******************************************************************************
*									      *
* Implemented instructions (to be performed on dataregisters only)	      *
*									      *
* fint		round off to integer value				      *
* fabs		absolute value						      *
* fmul		fixedpoint multiply					      *
* fdiv		fixedpoint divide					      *
* fsin		sine							      *
* fcos		cosine							      *
* ftan		tangent							      *
* fatan		arctan							      *
* fetox		base 2 exponentiation 					      *
* fsqrt		square root						      *
*									      *
*									      *
* Supports the same addressing modes as muls.w/divs.w			      *
*									      *
*******************************************************************************


; Constants

pi		=	(314159<<8)/100000
e		=	(271828<<8)/100000



*******************************************************************************
*									      *
*	fint dn								      *
*									      *
*	Return the magnitude of dn (8.8), signed			      *
*									      *
*	No registers modified						      *
*									      *
*******************************************************************************
		macro	fint	arg
		asr.w	#8,\arg

		endm


*******************************************************************************
*									      *
*	fabs dn								      *
*									      *
*	Return the absolute value of a number				      *
*									      *
*	No registers modified						      *
*									      *
*******************************************************************************
		macro   fabs	arg
.\~abs:		neg.w	\arg
		bmi.s	.\~abs

		endm


*******************************************************************************
*									      *
*	fmul dx,dy							      *
*									      *
*	Multiply two fixed point numbers (8.8), signed			      *
*									      *
*	Returns a 24.8 bit result					      *
*	No registers modified						      *
*									      *
*******************************************************************************
		macro	fmul	arg1,arg2
		muls.w	\arg1,\arg2
		asr.l	#8,\arg2

		endm


*******************************************************************************
*									      *
*	fdiv dx,dy							      *
*									      *
*	Divide dx (24.8) by dy (8.8), signed				      *
*									      *
*	Returns an 8.8 bit result					      *
*	No registers modified						      *
*									      *
*******************************************************************************
		macro	fdiv	arg1,arg2
		asl.l	#8,\arg2
		divs.w	\arg1,\arg2

		endm


*******************************************************************************
*									      *
*	fsin dn								      *
*									      *
*	Compute sine of dn (fineangle range is 16 bits)			      *
*									      *
*	Returns an 8.8 bit result					      *
*	No registers modified						      *
*									      *
*	Pade approximation:						      *
*	  sin(x) = x(60-7x)/(60+3x)					      *
*									      *
*******************************************************************************
		macro	fsin	arg

; Determine quadrant and clamp angle

		bclr.l	#14,\arg	; Even or odd quadrant?
		beq.s	.\~even		; Even

		eori.w	#$3fff,\arg	; Reverse the curve otherwise

.\~even:	bclr.l	#15,\arg	; Which semicircle?
		beq.s	.\~first	; First

		neg.w	\arg		; Mirror curve against x-axis
					; on the second one
.\~first:

; Approximate

		muls.w	#2*pi,\arg	; Translate into radians
		swap.w	\arg

		move.w	\arg,-(sp)	; x
		fmul	\arg,\arg
		move.w	\arg,-(sp)	; x

		asl.w	#3,\arg		; 8x
		sub.w	(sp),\arg	; 7x
		neg.w	\arg
		addi.w	#60<<8,\arg	; 60-7x
		muls.w	2(sp),\arg	; x(60-7x)	16.16
		move.l	\arg,-(sp)

		move.w	4(sp),\arg	; x
		add.w	\arg,\arg	; 2x
		add.w	4(sp),\arg	; 3x
		addi.w	#60<<8,\arg	; 60+3x
		move.w	\arg,-(sp)
		move.l	2(sp),\arg
		divs.w	(sp)+,\arg	; x(60-7x)/(60+3x)

		addq.l	#8,sp
		endm


*******************************************************************************
*									      *
*	fcos dn								      *
*									      *
*	Compute cosine of dn (fineangle range is 16 bits)		      *
*									      *
*	Returns an 8.8 bit result					      *
*	No registers modified						      *
*									      *
*******************************************************************************
		macro	fcos	arg
		addi.w	#$3fff,\arg	; cos(x) = sin(x+/2)
		fsin	\arg

		endm


*******************************************************************************
*									      *
*	ftan dn								      *
*									      *
*	Compute tangent of dn (fineangle range is 16 bits)		      *
*									      *
*	Returns an 8.8 bit result					      *
*	No registers modified						      *
*									      *
*******************************************************************************
		macro	ftan	arg
		move.w	\arg,-(sp)
		fcos	\arg
		move.w	\arg,-(sp)

		move.w	2(sp),\arg
		fsin	\arg

		tst.w	(sp)
		bne.s	.\~compute	; Catch borderline cases with cos(x)=0

		fmul	#$7fff,\arg	; Return +/-
		bra.s	.\~exit

.\~compute:	ext.l	\arg
		fdiv	(sp),\arg	; tan(x) = sin(x)/cos(x)

.\~exit:	addq.w	#4,sp

		endm


*******************************************************************************
*									      *
*	fatan dn							      *
*									      *
*	Compute arctan of the given argument				      *
*									      *
*	Returns an 8.8 bit result					      *
*	No registers modified						      *
*									      *
*	Pade approximation:						      *
*	  x(15+4x)/(15+9x)						      *
*									      *
*******************************************************************************
		macro	fatan	arg

; Preadjust

		clr.w	-(sp)		; adj = 0

		cmpi.w	#$0100,\arg
		ble.s	.\~below

		move.w	\arg,-(sp)
		move.l	#$ffff0000,\arg
		divs.w	(sp)+,\arg	; x = -1/x

		move.w	#(pi)/2,(sp)	; adj = /2

.\~below:	cmpi.w	#$ff00,\arg
		bgt.s	.\~above

		move.w	\arg,-(sp)
		move.l	#$ffff0000,\arg
		divs.w	(sp)+,\arg	; x = -1/x

		move.w	#-((pi)/2),(sp)	; adj = -/2

; Approximate

.\~above:	move.w	\arg,-(sp)	; x
		fmul	\arg,\arg
		move.w	\arg,-(sp)	; x
		add.w	\arg,\arg
		add.w	\arg,\arg	; 4x
		move.w	\arg,-(sp)

		addi.w	#15<<8,\arg
		move.w	\arg,-(sp)	; 15+4x
		fmul	6(sp),\arg
		move.l	\arg,-(sp)	; x(15+4x)

		move.w	4(sp),\arg
		add.w	6(sp),\arg
		add.w	8(sp),\arg
		move.w	\arg,-(sp)	; 15+9x

		move.l	2(sp),\arg
		fdiv	(sp)+,\arg	; x(15+4x)/(15+9x)

		lea.l	12(sp),sp

		add.w	(sp)+,\arg	; Add adjustor

		endm


*******************************************************************************
*									      *
*	fetox dx,dy							      *
*									      *
*	Compute exp2 to dx (8.8), unsigned				      *
*									      *
*	Returns a 24.8 bit result in dy					      *
*	No registers modified						      *
*									      *
*	Pade approximation:						      *
*	  (3+x)/(3-x)							      *
*									      *
*******************************************************************************
		macro	fetox	arg,res
		move.w	\arg,-(sp)

		moveq.l	#1,\res		; Process magnitude by bitshifting
		lsr.w	#8,\arg
		lsl.w	\arg,\res	; 2^int(x) = 1<<int(x)

		move.w	(sp),\arg	; Restore x
		move.w	\res,-(sp)

		andi.w	#$ff,\arg	; Process fraction using approximation
		move.w	\arg,\res
		neg.w	\arg
		addi.w	#3<<8,\arg	; (3-x)
		addi.w	#3<<8,\res	; (3+x)
		ext.l	\res
		fdiv	\arf,\res	; 2^frac(x) ~ (3+x)/(3-x)

		mulu.w	(sp)+,\res	; 2^x = 2^(int(x)+frac(x))
					;     = 2^int(x)*2^frac(x)
		move.w	(sp)+,\arg

		endm


*******************************************************************************
*									      *
*	fsqrt dx,dy							      *
*									      *
*	Compute squareroot of dx (16.8)					      *
*									      *
*	Returns a 24.8 bit result in dy					      *
*	No registers modified						      *
*									      *
*******************************************************************************
		macro	fsqrt	arg,res
		move.l	\arg,-(sp)
		asl.l	#8,\arg		; Shift to 16.16 precision
		move.l	\arg,\res	; b=a=r

.\~iterate:	move.l	(sp),\arg
		divu.w	\res,\arg	; b=r/a
		ext.l	\arg
		add.l	\arg,\res	; a=(b+a)/2
		asr.l	#1,\res

		cmp.l	\arg,\res
		bgt.s	.\~iterate	; Computing with signed numbers here in
					; order to allow fast longword extension
		move.l	(sp)+,\arg

        	endm
        	